from pytket.circuit import Circuit
from pyquil import Program
import pyquil.gates as G
from sympy import Symbol
import pennylane as qml
import cirq as cirq
import qiskit as qk

from compiler import toCirq, toQiskit, toTket, toPenny, toCudaq
from parser import toQasm

dev = qml.device("default.qubit", wires=2)
PARSE = False

if PARSE:
  @qml.qnode(dev)
  def bell_penny(angles):

    qml.Hadamard(wires=1)
    qml.CNOT(wires=[0, 1])
    qml.RY(angles[0], wires=0)
    qml.RY(angles[1], wires=1)
    qml.RY(angles[2], wires=1)
    qml.RY(angles[3], wires=0)

    return qml.state()

  def bell_ibm():
    from qiskit.circuit import Parameter, QuantumCircuit
    a,b = Parameter('a'), Parameter('b')
    qc = QuantumCircuit(2)
    qc.h(0)
    qc.cx(0, 1)
    qc.ry(a, 0)
    qc.ry(b, 1)
    qc.measure_all()

    return qc

  def bell_quan():
    theta1, theta2 = Symbol('theta1'), Symbol('theta2')
    circ = Circuit(2)
    circ.H(0)
    circ.CX(0, 1)
    circ.Ry(theta1, 0)
    circ.Ry(theta2, 1)

    return circ

  def bell_quil():
    p = Program()
    ro = p.declare("ro", "BIT", 2)
    theta1 = p.declare("theta1", "REAL")
    theta2 = p.declare("theta2", "REAL")

    p += G.H(0)
    p += G.CNOT(0, 1)
    p += G.RY(theta1, 0)
    p += G.RY(theta2, 1)
    p += G.MEASURE(0, ro[0])
    p += G.MEASURE(1, ro[1])

    return p

  def bell_cirq():
    from sympy import Symbol
    theta1, theta2 = Symbol('theta1'), Symbol('theta2')

    circ = cirq.Circuit()
    q0, q1 = cirq.LineQubit.range(2)
    circ.append(cirq.H(q0))
    circ.append(cirq.CNOT(q0, q1))
    circ.append(cirq.ry(theta1)(q0))
    circ.append(cirq.ry(theta2)(q1))

    return circ

  def bell_cudaq():
    kernel = make_kernel()
    q = kernel.qalloc(q)

    kernel.h(q[0])
    kernel.cx(q[0], q[1])
    kernel.rx(0.1*0, q[0])
    kernel.rx(0.1*1, q[1])

    kernel.mz(q[0])
    kernel.mz(q[1])

    return kernel
  # print(toQasm(bell_penny))
  # print(toQasm(bell_quan()))
  # print(toQasm(bell_ibm()))
  # print(toQasm(bell_cirq()))
  # print(toQasm(bell_quil()))
  # print(toQasm(bell_cudaq()))


QASM="""
OPENQASM 2.0;
include "qelib1.inc";

qreg q[2];


h q[0];
cx q[0],q[1];
ry(var_theta1) q[0];
ry(var_theta2) q[1];
""".strip()

# print(toQiskit(QASM))
print(qml.draw(toPenny(QASM, dev))())

# print(toCirq(QASM))
# print(toTket(QASM))
# print(toCudaq(QASM))


from re import compile, findall
import numpy as np

vargex = compile("var_\w*")

def autoParam(int):
  rand = np.round(np.random.rand(), 6)
  rand = [rand+(i/1e5) for i in range(int)]
  rand = np.round(rand, 7)
  return rand

def parseVars(string):
  variables = []

  # string will contain var_0, var_1, etc.
  for line in string.split("\n"):
    if "var_" in line:
      matches = findall(vargex, line)
      variables.extend(matches)
    # endif
  # endfor
  variables = list(set(variables))
  params = list(autoParam(len(variables)))

  for i, var in enumerate(variables):
    string = string.replace(var, str(params[i]))

  variables = [i.replace('var_', '') for i in variables]
  return string, variables, params


def toQiskit(string):
  from qiskit.circuit import QuantumCircuit, Parameter
  from qiskit.transpiler import TransformationPass
  from qiskit.converters import circuit_to_dag

  string, variables, params = parseVars(string)
  variables = [Parameter(p) for p in variables]

  qc = QuantumCircuit.from_qasm_str(string)
  class BackReplace(TransformationPass):
    def run(self, dag):
      for node in dag.op_nodes():
        cparams = node.op.params
        if len(cparams) == 0:
          continue
        replacement = QuantumCircuit(len(cparams))
        op = node.op.copy()
        new_params = []
        for p in cparams:
          if p in params:
            idx = params.index(p)
            new_params.append(variables[idx])
          else:
            new_params.append(p)

        op.params = new_params
        replacement.append(op, [0])

        dag.substitute_node_with_dag(node, circuit_to_dag(replacement))
      return dag

  transpiled = BackReplace()(qc)
  return transpiled

def toPenny(string, device):
  from pennylane import qml
  from pennylane.tape import make_qscript

  string, variables, params = parseVars(string)
  qc = qml.from_qasm(string)
  ops = list(make_qscript(qc)())
  new_ops = []
  for op in ops:
    if op.num_params > 0:
      new_params = []
      for p in op.parameters:
        if p in params:
          idx = params.index(p)
          p = variables[idx]
        # endif
        new_params.append(p)
      # endfor
      op.data = tuple(
        np.array(p)
          if isinstance(p, (list, tuple))
          else p for p in new_params
      )
    # endif
    new_ops.append(op)
  # endfor

  def circuit():
    for op in new_ops:
      qml.apply(op)
    return qml.state()

  circuit=qml.QNode(circuit, device)
  return circuit

"""START OF TKET"""
def printket(circ):
  from pytket.circuit.display import get_circuit_renderer
  CR = get_circuit_renderer()
  CR.set_render_options(zx_style=True)
  CR.condense_c_bits = False
  CR.min_height = "300px"
  print(CR.render_circuit_as_html(circ))

def unstring(pars):
  pars = [i.replace('/pi', '') for i in pars]
  pars = [i[1:-1] if i[0] == '(' else i for i in pars]
  pars = [float(i) for i in pars]
  return pars

# use Circuit.to_dict | Circuit.from_dict for debugging
def toTket(string):
  from pytket.qasm.qasm import parser, CircuitTransformer as CT
  from pytket import Circuit

  string, variables, params = parseVars(string)
  jsond = parser(maxwidth=32).parse(string)

  for cmd in jsond['commands']:
    if 'params' not in cmd['op']:
      continue # no parameters
    cpars = unstring(cmd['op']['params'])
    new_params = []
    for p in cpars:
      if p in params:
        p = variables[params.index(p)]
      new_params.append(p)
    # endfor
    cmd['op']['params'] = new_params
  # endfor

  qc = Circuit.from_dict(jsond)
  return qc

def toCirq(string):
  from cirq.contrib.qasm_import import circuit_from_qasm
  from sympy import Symbol
  import cirq as c
  string, variables, params = parseVars(string)

  def map_func(op, _):
    if hasattr(op.gate, 'exponent'):
      expo = op.gate.exponent
      if expo in params:
        idx = params.index(expo)
        idx = Symbol(str(variables[idx]))

        gate = getattr(c, op.gate.__class__.__name__)
        op = gate(rads=idx).on(*op.qubits)
      # endif
    yield op

  params = [i/np.pi for i in params]
  # print(string, variables, params)

  qc = circuit_from_qasm(string)
  qc2 = c.map_operations(qc, map_func)

  return qc2

def toCudaq(string):
  from qiskit.transpiler.passes import RemoveBarriers
  from qiskit.circuit import Parameter
  from qiskit import QuantumCircuit
  from cudaq import make_kernel

  string, variables, params = parseVars(string)
  variables = [Parameter(p) for p in variables]

  qc = QuantumCircuit.from_qasm_str(string)
  print(string)
  print(qc)
  qc = RemoveBarriers()(qc)

  if len(params) > 0:
    kernel, thetas = make_kernel(list)
  else:
      kernel = make_kernel()
  cuq = kernel.qalloc(len(qc.qubits))

  def u(self, a, b, c, u):
    kernel.rz(b, u)
    kernel.ry(a, u)
    kernel.rz(c, u)

  kernel.u = u
  kernel.p = kernel.r1
  gate_map = {
    "phase": "p", "swap": "swap","measure": "mz"
  }
  legal_gates = [
    'x', 'y', 'z', 'phase',
    'rx', 'ry', 'rz', 'cx', 'cz', 'cy',
    'u', 's', 't', 'h', 'swap', 'measure'
  ]
  # just loop over the gates and apply them

  for instr, qubits, _ in qc.data:
    if instr.name not in legal_gates:
      raise ValueError(f"Gate {instr.name} not supported")
    name = instr.name
    arguments = instr.params
    if name in gate_map:
      name = gate_map[name]

    op = getattr(kernel, name)
    qubits = [q._index for q in qubits]
    qubits = [cuq[q] for q in qubits]
    if len(arguments) > 0:
      # arguments need to be mapped to thetas if theyre in params
      new_args = []
      for arg in arguments:
        if arg in params:
          idx = params.index(arg)
          arg = thetas[idx]
        new_args.append(arg)

      arguments = new_args
      op(*new_args, *qubits)
    else:
      op(*qubits)

  return kernel


from setuptools import setup, find_packages

with open('Readme.md', 'r') as fh:
  long_description = fh.read()

setup(
  name='abrax',
  version='0.5.0',
  description='Cross platform circuit transpiler',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author='plutoniumm',
  url='https://github.com/plutoniumm/abraxas',
  license='MIT',
  package_dir={'': '.'},
  packages=find_packages(where='.'),
  python_requires='>=3.9',
  install_requires=[ 'numpy' ],
  extras_require={
    'dev': ['pytest', 'twine', 'wheel'],
  },
)



import numpy as np
import re

class Unlist(list):
  def __init__(self):
    self.append(0)
    self.__class__.__module__ = "builtins"

  def __getitem__(self, index):
    while index >= len(self):
      self.append(index)
    return super(Unlist, self).__getitem__(index)

def compile_penny(qc, params):
  qc.construct([params], {})
  t = qc.qtape.to_openqasm().strip()

  return t

def compile_qiskit(qc, params):
  from qiskit.qasm2 import dumps
  qc = qc.assign_parameters(params)
  t = dumps(qc).strip()

  return t

def compile_tket(qc, params):
  from qiskit.qasm2 import dumps
  from pytket.extensions.qiskit import tk_to_qiskit
  params = params / np.pi
  qc.symbol_substitution(
    dict(zip(qc.free_symbols(), params))
  )
  t = tk_to_qiskit(qc)
  t = dumps(t).strip()

  return t

def compile_cirq(qc, params):
  from cirq import qasm, resolve_parameters,ParamResolver
  names = list(qc._parameter_names_())
  params = params * np.pi
  qc = resolve_parameters(
    qc, ParamResolver(dict(zip(names, params)))
  )
  t = qasm(qc).strip()

  return t

"""START: compile_quil"""
def _quil_parse_decl(line):
  line = line.split(" ")
  typ = line[2] # BIT[2]
  size = int(typ.split("[")[1].split("]")[0])

  typ = typ.split("[")[0]

  return typ, size

def _quil_parse_gate(line):
  line = line.split(" ")
  qubit = line[-1]
  full_instr = line[0] # RY(theta1[0]) OR H
  instr = full_instr.split("(")[0]
  if "(" in full_instr:
    params = full_instr.split("(")[1].split(")")[0]
    params = params.split("[")[0].split(",")
    # now we need to check if theres letters in the params
    # if yes we prefix with 'var_'
    for i in range(len(params)):
      if any(c.isalpha() for c in params[i]):
        params[i] = "var_" + params[i]
  else:
    params = []

  if len(line) == 3:
    qubit = [line[1], qubit]

  return instr, qubit, params

def compile_quil(quilstr):
  quilstr = re.sub(r"//.*\n", "", quilstr)
  init = "OPENQASM 2.0;\ninclude \"qelib1.inc\";"
  qasm = ""

  # qregs, cregs, variables
  qubits = []
  variables = []
  for line in quilstr.split("\n"):
    if "DECLARE" in line:
      typ, size = _quil_parse_decl(line)
      if typ == "BIT":
        qasm += f"creg meas[{size}];\n"
    elif "MEASURE" in line:
      [_, qubit, creg] = line.split(" ")
      creg = int(creg.split("[")[1].split("]")[0])
      qasm += f"measure q[{qubit}] -> meas[{creg}];\n"
      qubits.append(qubit)
    else:
      instr, qubit, params = _quil_parse_gate(line)
      if len(instr) < 1 or len(qubit) < 1:
        continue
      variables.extend(params)
      if len(params) > 0:
        gate = f"{instr.lower()}({params[0]}) q[{qubit}]"
        qubits.append(qubit)
      else:
        if len(qubit) == 2:
          gate = f"{instr.lower()} q[{qubit[0]}],q[{qubit[1]}]"
          qubits.append(qubit[0])
          qubits.append(qubit[1])
        else:
          gate = f"{instr.lower()} q[{qubit}]"
          qubits.append(qubit)
      # endif
      qasm += f"{gate};\n"
    # endif
  # endfor
  qubits = list(set(qubits))
  qasm = qasm.replace("cnot", "cx")
  qasm = "\nqreg q[{}];\n".format(len(qubits)) + qasm
  qasm = (init + qasm).strip()

  variables = list(set(variables))
  variables = [i for i in variables if i.startswith("var_")]
  return qasm, variables

def autoParam(int):
  rand = np.round(np.random.rand(), 6)
  rand = [rand+(i/1e5) for i in range(int)]
  rand = np.round(rand, 7)
  return rand

def toQasm(qc):
  name = qc.__class__.__name__
  base = qc.__class__.__base__.__name__
  params = [0]
  qasmd = None

  if name == 'QuantumCircuit': # Qiskit
    names = [i.name for i in qc.parameters]
    params = autoParam(len(names))
    qasmd = compile_qiskit(qc, params=params)

  elif name == 'QNode': # PennyLane
    qc2 = qc
    names = Unlist()
    qc2(names)
    names = [str(i) for i in names]
    params = autoParam(len(names))
    qasmd = compile_penny(qc2, params=params)

  elif base == 'pybind11_object': # tket
    names = []
    for p in qc:
      names.extend(p.free_symbols())
    names = list(set(names))
    params = autoParam(len(names))
    qasmd = compile_tket(qc, params=params)

  elif name == 'Circuit': # Cirq
    names = list(qc._parameter_names_())
    params = autoParam(len(names))
    qasmd = compile_cirq(qc, params=params)

  elif name == 'Program': # Quil
    qasmd, names = compile_quil(qc.out())
    params = autoParam(len(names))
  elif name == 'PyKernel': # CudaQ
    if hasattr(qc, 'arguments'):
      args = qc.arguments
      if len(args) > 0:
        raise ValueError('Arguments not supported yet.')

    from cudaq import translate
    params = []
    qasmd = translate(qc, format="openqasm2")
  else:
    raise ValueError(f'Unsupported circuit: {name}')

  # replace all params back with var_i
  for i in range(len(params)):
    qasmd = qasmd.replace(str(params[i]), f'var_{names[i]}')

  return qasmd